#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <readline/readline.h>
#include <readline/history.h>

#define ASIZE sizeof(struct sockaddr_in)

char *pname;
unsigned short game_port = 1337;
char game_ip[16] = "0.0.0.0";
int game_fd;

const char* rl_readline_name = "NavBat";

void process_command(char* line)
{
  if(line==NULL) exit(0);
  if(strlen(line)==0) return;
  add_history(line);

  int n = strlen(line);
  char* buf = malloc((n+2)*sizeof(char));
  sprintf(buf, "%s\n", line);
  send(game_fd, buf, n+1, 0);
  free(buf);
}

void parse_line(char* buf)
{
 int i; char name[65];
 for(i=0; i<strlen(rl_line_buffer);i++) printf("\b \b");

 if(strncmp(buf, "/msg ", 5)==0)
 {
  sscanf(buf+5, "%s", name);
  if(strcmp(name, "server")==0)
   printf(" *** \e[1m%s\e[0m", buf+12);
  else
   printf("%s : %s", name, buf+6+strlen(name));
 }
 else if(strncmp(buf, "/error", 6)==0)
 {
  printf(" !!! \e[4m%s\e[0m", buf+7);
 }
 else
 {
  printf(" +++ %s", buf);
 }

 fflush(stdout);
 rl_forced_update_display();
}

void client_loop()
{
 int fdmax, in_fd;
 int i, len;
 struct sockaddr_in game_addr;
 char buffer[1024];
 fd_set master_set, read_fds;

 if ((game_fd=socket(AF_INET,SOCK_STREAM,0))<0)
 {
  perror("Fatal error: unable to create game socket.");
  exit(1);
 }

 game_addr.sin_family = AF_INET;
 game_addr.sin_port = htons(game_port);
 game_addr.sin_addr.s_addr = inet_addr(game_ip);

 if(connect(game_fd, (struct sockaddr*) &game_addr, ASIZE) < 0)
 {
  perror("Fatal error: unable to join the game.");
  exit(1);
 }

 in_fd = fileno(stdin);
 FD_SET(game_fd, &master_set);
 FD_SET(in_fd, &master_set);
 fdmax = game_fd;

 rl_callback_handler_install("", process_command);

 while(1)
 {
  read_fds = master_set;

  if(select(fdmax+1, &read_fds, NULL, NULL, NULL) < 0)
  {
    perror("Warning: select() failed.");
    continue;
  }

  for(i=0; i <= fdmax; i++)
  {
    if(FD_ISSET(i, &read_fds))
    {
     if(i == game_fd)
     {
      if((len = recv(i, buffer, sizeof(buffer), 0)) <= 0)
      {
       if(len == 0)
        printf("Lost connection to host.\n");
       else
        perror("recv() error !\n");

       rl_callback_handler_remove();
       close(in_fd); close(game_fd);
       FD_ZERO(&master_set); FD_ZERO(&read_fds);
       exit(0);
      }
      else
      {
       buffer[len]=0;
       parse_line(buffer);
      }
     }
     else if(i == in_fd)
     {
      rl_callback_read_char();
     }
    }
   }
 }
}

void usage()
{
  printf("Usage : %s [options]\n\n", pname);
  printf("  A client for a network naval battle game.\n");
  printf("  Possible options for this command are :\n");
  printf("    -h, --help : display this help message\n");
  printf("    -c, --connect : set serveur IP (defaults to 127.0.0.1)\n");
  printf("    -p, --port : set port number (defaults to 1337)\n");
  exit(0);
}

void clerror(char *msg, char *more)
{
 printf("%s %s\n", msg, more);
 usage();
}

int main(int argc, char** argv)
{
 int i;
 char *s;

 pname = argv[0];

 for(i=1; i<argc; i++)
 {
  s=argv[i];
  if(strcmp(s, "-h")*strcmp(s,"--help") == 0)
  {
   usage();
  }
  else if(strcmp(s, "-p")*strcmp(s,"--port") == 0)
  {
    if(++i >= argc) clerror("Missing port number after option", s);
    game_port = (unsigned short) atoi(argv[i]);
    if(game_port <= 1024 || game_port > 65534) clerror("Invalid port number provided for option", s);
  }
  else if(strcmp(s, "-c")*strcmp(s,"--connect") == 0)
  {
    if(++i >= argc) clerror("Missing server IP after option", s);
    strncpy(game_ip, argv[i], 15);
    game_ip[15]=0;
    if(inet_addr(game_ip) < 0) clerror("Invalid IP address provided for option", s);
  }
  else
  {
   clerror("Unknown command line option:", s);
  }
 }

 client_loop();

 return 0;
}
