#include "server.h"
#define ASIZE sizeof(struct sockaddr_in)

char *pname;
unsigned short game_port = 1337;
unsigned short observer_port = 1408;
char game_bind[16] = "0.0.0.0";
char observer_bind[16] = "0.0.0.0";
int game_fd, observer_fd, fdmax;
GHashTable *players, *lobbies, *obs;
fd_set master_set;

void free_obs(observer *o)
{
 free(o);
}

void free_player(player *p)
{
 if(p->s)
 {
  if(p->s->grid) free(p->s->grid);
  free(p->s);
 }
 free(p);
}

void free_lobby(lobby *l)
{
 int i;
 for(i=0; i<l->nb_boats; i++)
 {
  free(l->boats[i]->shape);
  free(l->boats[i]);
 }
 free(l->boats); free(l->members);
 g_array_free(l->fires, TRUE);
 free(l);
}

void respond(int to, char *msg)
{
 if(send(to, msg, strlen(msg), 0) < 0)
  perror("send() error");
}

void broadcast(char *msg, player *scope, bool include)
{
 void sendto(int *k, player *p, void *n)
 {
  if((scope==NULL && (include || p->lobby_id==0))
   || (scope != NULL && p->lobby_id == scope->lobby_id && (include || p!=scope)))
   respond(p->fd, msg);
 }
 g_hash_table_foreach(players, (GHFunc) sendto, NULL);
}

int leave_game(player *p, lobby *l)
{
 char msg[1024];
 int i; bool end=l->connected <= 2;
 if(p)
 {
  p->st = DEFAULT;
  p->lobby_id = 0;
  free(p->s->grid);
  free(p->s);
  p->s = NULL;
  sprintf(msg, "/msg server %s left the game %s.\n", p->name, l->name);
 }

 for(i=0; i<l->connected; i++)
 {
  if(p != NULL)
  {
   respond(l->members[i]->fd, msg);
   if(l->members[i] == p)
   {
    if(l->st == INGAME && i == l->turn){
     l->turn = (l->turn + 1) % (l->connected);
     respond(l->members[l->turn]->fd, "/play\n");
    }
    l->members[i]=l->members[--l->connected];
   }
  }
  if(p==NULL || (end && l->members[i] != p))
  {
   respond(l->members[i]->fd, (p==NULL ? "/endgame\n" : "/msg server You win the game !\n"));
   l->members[i]->st = DEFAULT;
   l->members[i]->lobby_id = 0;
   free(l->members[i]->s->grid);
   free(l->members[i]->s);
   l->members[i]->s = NULL;
  }
 }
 return (p==NULL || end);
}

void desert(player *p)
{
 if(p->st != DEFAULT)
 {
  int *lobid = malloc(sizeof(int));
  *lobid = p->lobby_id;
  if(leave_game(p, g_hash_table_lookup(lobbies, lobid)))
  {
   g_hash_table_remove(lobbies, lobid);
  }
 }
}

void invite_place(player *p, boat *b)
{
 char msg[2048], tmp[128];
 int i,j;

 sprintf(msg, "/msg server Please place the following boat :\n");
 for(i=0; i<b->height; i++)
 {
  for(j=0; j<b->width; j++) tmp[j]=(b->shape[i*b->width+j] ? '*' : ' ');
  tmp[j]=0;
  sprintf(msg, "%s\t%s\n", msg, tmp);
 }
 respond(p->fd, msg);
}

void init_state(player *p, lobby *l)
{
 p->s = malloc(sizeof(pst));
 p->s->cur_boat = 0;
 p->s->alive_boats = l->nb_boats;
 p->s->grid = calloc(l->dimensions[0]*l->dimensions[1]*l->dimensions[2],sizeof(int));
}

void server_loop()
{
 int i, j, k, newfd, len, lobid = 0;
 struct sockaddr_in game_addr, observer_addr, client_addr;
 struct timeval tv;
 socklen_t client_size;
 fd_set read_fds;
 char buffer[1025], tmp[1500];
 player *cur_player;
 lobby *cur_lobby;
 observer *o;
 int *cur_index;

 tv.tv_sec = 1;
 tv.tv_usec = 0;

 if ((game_fd=socket(AF_INET,SOCK_STREAM,0))<0)
 {
  perror("Fatal error: unable to create game socket");
  exit(1);
 }
 if ((observer_fd=socket(AF_INET,SOCK_STREAM,0))<0)
 {
  perror("Fatal error: unable to create observer socket");
  exit(1);
 }

 game_addr.sin_family = AF_INET;
 game_addr.sin_port = htons(game_port);
 game_addr.sin_addr.s_addr = inet_addr(game_bind);

 observer_addr.sin_family = AF_INET;
 observer_addr.sin_port = htons(observer_port);
 observer_addr.sin_addr.s_addr = inet_addr(observer_bind);

 if(bind(game_fd, (struct sockaddr*) &game_addr, ASIZE) < 0)
 {
  perror("Fatal error: unable to bind game socket");
  exit(1);
 }

 if(bind(observer_fd, (struct sockaddr*) &observer_addr, ASIZE) < 0)
 {
  perror("Fatal error: unable to bind observer socket");
  exit(1);
 }

 if(listen(game_fd, 10) < 0)
 {
  perror("Fatal error: cannot listen on game socket");
  exit(1);
 }

 if(listen(observer_fd, 10) < 0)
 {
  perror("Fatal error: cannot listen on observer socket");
  exit(1);
 }

 FD_SET(game_fd, &master_set);
 FD_SET(observer_fd, &master_set);
 fdmax = game_fd > observer_fd ? game_fd : observer_fd;
 players = g_hash_table_new_full(g_int_hash, g_int_equal, free, (GDestroyNotify) free_player);
 lobbies = g_hash_table_new_full(g_int_hash, g_int_equal, free, (GDestroyNotify) free_lobby);
 obs = g_hash_table_new_full(g_int_hash, g_int_equal, free, (GDestroyNotify) free_obs);

 while(1)
 {
  read_fds = master_set;

  if(select(fdmax+1, &read_fds, NULL, NULL, NULL) < 0)
  {
    perror("Warning: select() failed");
    continue;
  }

  for(i=0; i <= fdmax; i++)
  {
    if(FD_ISSET(i, &read_fds))
    {
     if(i == game_fd)
     {
      client_size = sizeof(client_addr);
      memset(&client_addr, 0, client_size);
      if((newfd = accept(game_fd, (struct sockaddr *)&client_addr, &client_size)) < 0)
      {
       perror("Cannot accept game client");
      }
      else
      {
       FD_SET(newfd, &master_set);
       if(newfd > fdmax)
       {
        fdmax = newfd;
       }
       printf("Connection from %s\n", inet_ntoa(client_addr.sin_addr));

       cur_player = malloc(sizeof(player));
       sprintf(cur_player->name, "player_%d", newfd);
       cur_player->addr = client_addr;
       cur_player->fd = newfd;
       cur_player->st = DEFAULT;
       cur_player->lobby_id = 0;
       cur_player->s = NULL;
       cur_index = malloc(sizeof(int));
       *cur_index = newfd;

       g_hash_table_insert(players, cur_index, cur_player);

       respond(newfd, "/msg server Welcome to this game server!\n");
       sprintf(tmp, "/msg server %s joined the server\n", cur_player->name);
       broadcast(tmp, cur_player, false);
      }
     }
     else if(i == observer_fd)
     {
      client_size = sizeof(client_addr);
      memset(&client_addr, 0, client_size);
      if((newfd = accept(game_fd, (struct sockaddr *)&client_addr, &client_size)) < 0)
      {
       perror("Cannot accept game client");
      }
      else
      {
       FD_SET(newfd, &master_set);
       if(newfd > fdmax)
       {
        fdmax = newfd;
       }
       printf("Observer connection from %s\n", inet_ntoa(client_addr.sin_addr));

       o = malloc(sizeof(observer));
       sprintf(o->name, "observer_%d", newfd);
       o->addr = client_addr;
       o->fd = newfd;
       o->port = 8080;
       o->lobby_id = 0;
       cur_index = malloc(sizeof(int));
       *cur_index = newfd;

       g_hash_table_insert(obs, cur_index, o);

       respond(newfd, "/msg server Welcome to this game server!\n");
      }
     }
     else if((cur_player = g_hash_table_lookup(players, &i)))
     {
      cur_lobby = g_hash_table_lookup(lobbies, &cur_player->lobby_id);

      if((len = recv(i, buffer, sizeof(buffer), 0)) <= 0)
      {
       if(len == 0)
        printf("Client %d disconnected.\n", i);
       else
        perror("recv() error !\n");

       desert(cur_player);
       g_hash_table_remove(players, &i);
       close(i);
       FD_CLR(i, &master_set);
      }
      else if(strncmp(buffer, "/quit", 5)==0)
      {
       printf("Client %d left.\n", i);
       sprintf(tmp, "/msg server %s left the game\n", cur_player->name);
       broadcast(tmp, cur_player, false);

       desert(cur_player);
       g_hash_table_remove(players, &i);
       close(i);
       FD_CLR(i, &master_set);
      }
      else if(strncmp(buffer, "/desert", 7)==0)
      {
       if(cur_player->st != DEFAULT)
        desert(cur_player);
       else respond(i, "/error You are not in any game or lobby.\n");
      }
      else if(strncmp(buffer, "/fire ", 6)==0)
      {
       bool alive, d3 = cur_lobby->nb_dim == 3;
       us id, x, y, z;
       int W, H, Z, u, v, w, h, n, r, *plane;
       W = cur_lobby->dimensions[2];
       H = cur_lobby->dimensions[1];
       Z = cur_lobby->dimensions[0];

       if(cur_player->st == INGAME && cur_lobby->members[cur_lobby->turn]==cur_player)
       {
        if((d3 && sscanf(buffer+6, "%hu %hu %hu %hu", &id, &x, &y, &z)==4)
         || ((z=1) && sscanf(buffer+6, "%hu %hu %hu", &id, &x, &y)==3))
        {
         if(id>=0 && id < cur_lobby->connected && cur_lobby->members[id] != cur_player)
         {
          if(z>=1 && z<=Z && x>=1 && x<=H && y>=1 && y <= W)
          {
           n = H*W; plane = (cur_lobby->members[id]->s->grid)+n*(z-1);
           r = plane[(x-1)*W+(y-1)];
           if(r>0)
           {
            plane[(x-1)*W+(y-1)] = -1*r; alive=false;
            w = cur_lobby->boats[r]->width;
            h = cur_lobby->boats[r]->height;
            for(u=x-h; u<x+h; u++)
             for(v=y-w; v<y+w; v++)
             {
              if(u>=0 && u<H && v>=0 && v<W && plane[u*W+v]==r) alive=true;
             }
            if(!alive)
            {
             if(--cur_lobby->members[id]->s->alive_boats==0)
             {
              sprintf(tmp, "/msg server %s has lost all his boats! Better luck next time...\n", cur_lobby->members[id]->name);
              broadcast(tmp, cur_player, true);
              alive = cur_lobby->connected > 2;
              desert(cur_lobby->members[id]);
              if(!alive) continue;
             }
            }
           }
           respond(i, "/msg server End of your turn.\n");
           cur_lobby->turn = (cur_lobby->turn + 1) % (cur_lobby->connected);
           respond(cur_lobby->members[cur_lobby->turn]->fd, "/play It is your turn to play.\n");
          }
          else respond(i, "/error Invalid target coordinates.\n");
         }
         else respond(i, "/error Invalid target ID.\n");
        }
        else respond(i, "/error Wrong arguments to the command.\n");
       }
       else respond(i, "/error It is not your turn to play yet.\n");
      }
      else if(strncmp(buffer, "/place ", 7)==0)
      {
       if(cur_player->st == INLOBBY)
       {
        boat *b = cur_lobby->boats[cur_player->s->cur_boat];
        char dir; bool d3;
        us x, y, z, w, h, W, H, Z;
        d3 = cur_lobby->nb_dim == 3;
        w = b->width; h=b->height;
        W = cur_lobby->dimensions[2];
        H = cur_lobby->dimensions[1];
        Z = cur_lobby->dimensions[0];

        if((d3 && sscanf(buffer+7, "%c %hu %hu %hu", &dir, &x, &y, &z)==4)
         || ((z=1) && sscanf(buffer+7, "%c %hu %hu", &dir, &x, &y)==3))
        {
         if((dir=='N' || dir=='S' || dir=='E' || dir=='W')
          && (x>=1 && x<=H) && (y>=1 && y<=W) && (z>=1 && z<=Z))
         {
          if(((dir=='N' || dir=='S') && x+h <= H && y+w <= W)
           || (((dir=='E' || dir=='W') && x+w <= H && y+h <= W)))
          {
           int u, v, x1, y1, n = cur_lobby->dimensions[1]*cur_lobby->dimensions[2];
           int *new_grid = malloc(n*sizeof(int));
           memcpy(new_grid, cur_player->s->grid+(n*(z-1)), n*sizeof(int));
           bool success = true; x1=y1=0;
           for(u=0; u<h; u++)
            for(v=0; v<w; v++)
            {
             switch(dir)
             {
              case 'N': x1=u; y1=v; break;
              case 'S': x1=(h-u-1); y1=v; break;
              case 'E': x1=(w-v-1); y1=u; break;
              case 'W': x1=v; y1=(h-u-1); break;
             }
             if(b->shape[u*w+v])
             {
              if(new_grid[(x-1+x1)*W+(y-1+y1)] != 0) success = false;
              new_grid[(x-1+x1)*W+(y-1+y1)] = 1+cur_player->s->cur_boat;
             }
            }
           if(success)
           {
/*
for(u=0; u<H; u++)
{
 for(v=0; v<W; v++)
  printf("%d ", new_grid[u*W+v]);
 printf("\n");
}
*/
            memcpy(cur_player->s->grid+(n*(z-1)), new_grid, n*sizeof(int));
            if(++cur_player->s->cur_boat < cur_lobby->nb_boats)
             invite_place(cur_player, cur_lobby->boats[cur_player->s->cur_boat]);
            else
            {
             cur_player->st = READY;
             for(j=0; j<cur_lobby->connected && cur_lobby->members[j]->st==READY; j++);
             if(j==cur_lobby->nb_players)
             {
              cur_lobby->st = GAME;
              for(j=0; j<cur_lobby->connected; j++) cur_lobby->members[j]->st=INGAME;
              cur_lobby->turn = 0;
              broadcast("/msg server All players ready, starting the game.\n", cur_player, true);
              respond(cur_lobby->members[0]->fd, "/play It is your turn to play.\n");
             }
             else respond(i, "/msg server You are ready to start the game.\n");
            }
           }
           else respond(i, "/error This position overlaps another boat.\n");
           free(new_grid);
          }
          else respond(i, "/error Out of bounds position given.\n");
         }
         else respond(i, "/error Invalid position : orientation and coordinates expected.\n");
        }
        else respond(i, "/error Wrong number of arguments.\n");
       }
       else respond(i, "/error Place your boats when entering a game lobby.\n");
      }
      else if(strncmp(buffer, "/join ", 6)==0)
      {
       int id;
       lobby *l;
       if(sscanf(buffer+6, "%d", &id)==1 && (l=g_hash_table_lookup(lobbies, &id)))
       {
        if(l->st == ACCEPT && l->connected < l->nb_players)
        {
         cur_player->st = INLOBBY;
         cur_player->lobby_id = id;
         init_state(cur_player, l);
         l->members[l->connected++] = cur_player;
         sprintf(tmp, "/join %d %s\n", i, cur_player->name);
         broadcast(tmp, cur_player, false);
         invite_place(cur_player, l->boats[cur_player->s->cur_boat]);
        }
        else respond(i, "/error This game is full or not ready.\n");
       }
       else respond(i, "/error Invalid game ID.\n");
      }
      else if(strncmp(buffer, "/create ", 8)==0)
      {
       unsigned short p;
       char n[128], rest[1024];
       rest[0]=0;

       if(cur_player->st != DEFAULT)
       {
        respond(i, "/error You must leave your current game with /desert first.\n");
       }
       else if((k=sscanf(buffer+8, "%64s %hu %1023[ 0-9]", n, &p, rest)) >= 2)
       {
        if(strlen(n) && p > 1 && k==2)
        {
         cur_lobby = malloc(sizeof(lobby));
         sprintf(cur_lobby->name, "%s", n);
         cur_lobby->nb_players = p;
         cur_lobby->connected = 1;
         cur_lobby->members = malloc(p * sizeof(player*));
         cur_lobby->members[0] = cur_player;
         cur_lobby->fires = g_array_new(FALSE, FALSE, sizeof(fire));
         cur_index = malloc(sizeof(int));
         *cur_index = ++lobid;

         if(k==3)
         {
          cur_player->st = CONFLOBBY;
         }
         else
         {
          cur_lobby->nb_boats = 1;
          cur_lobby->boats = malloc(sizeof(boat*));
          boat *b = cur_lobby->boats[0] = malloc(sizeof(boat));
          b->width = 1; b->height = 3;
          b->shape = malloc(3*sizeof(bool));
          b->shape[0]=b->shape[1]=b->shape[2]=true;
          cur_lobby->nb_dim = 2;
          cur_lobby->dimensions[0] = 1;
          cur_lobby->dimensions[1] = 10;
          cur_lobby->dimensions[2] = 10;
          cur_lobby->st = ACCEPT;
          cur_player->st = INLOBBY;
         }

         cur_player->lobby_id = lobid;
         init_state(cur_player, cur_lobby);
         g_hash_table_insert(lobbies, cur_index, cur_lobby);
         sprintf(tmp, "/id_game %d\n", lobid);
         respond(i, tmp);
         invite_place(cur_player, cur_lobby->boats[0]);
         sprintf(tmp, "/msg server %s created a new lobby : %s (#%d)\n", cur_player->name, n, lobid);
         broadcast(tmp, NULL, false);
        }
        else
        {
         respond(i, "/error Out of range values of arguments.\n");
        }
       }
       else respond(i, "/error Please provide at least a name and capacity.\n");
      }
      else if(strncmp(buffer, "/list players", 13)==0)
      {
       if(cur_player->st != DEFAULT)
       {
        cur_lobby = g_hash_table_lookup(lobbies, &cur_player->lobby_id);
        sprintf(tmp, "/msg server Players in %s :\n", cur_lobby->name);
        for(j=0; j < cur_lobby->connected; j++)
        {
         sprintf(tmp, "%s\t#%d\t%s\n", tmp, j, cur_lobby->members[j]->name);
        }
        respond(i, tmp);
       }
       else
       {
        sprintf(tmp, "/msg server Players on this server :\n");
        void splayer(int *k, player *p, void *n)
        {
         sprintf(tmp, "%s\t%s", tmp, p->name);
         if(p->lobby_id) sprintf(tmp, "%s\t In game #%d\n", tmp, p->lobby_id);
         else sprintf(tmp, "%s\n", tmp);
        }
        g_hash_table_foreach(players, (GHFunc)splayer, NULL);
        respond(i, tmp);
       }
      }
      else if(strncmp(buffer, "/list obs", 9)==0)
      {
       sprintf(tmp, "/msg server Name\tIP\tPort\tGame #\n");
       void addline(int *k, observer *o, void *n)
       {
        sprintf(tmp, "%s\t%s\t%s\t%d\t%d\n", tmp, o->name, inet_ntoa(o->addr.sin_addr), o->port, o->lobby_id);
       }
       g_hash_table_foreach(obs, (GHFunc) addline, NULL);
       respond(i, tmp);
      }
      else if(strncmp(buffer, "/list ", 6)==0)
      {
       if(cur_player->st == DEFAULT)
       {
        sprintf(tmp, "/msg server Current active lobbies :\n");
        void addline(int *k, lobby *l, void *n)
        {
         if((!strncmp(buffer+6, "waiting_games",13) && l->st == ACCEPT) || !strncmp(buffer+6, "games",5))
          sprintf(tmp, "%s\t#%d\t%s\t%d / %d\n", tmp, *k, l->name, l->connected, l->nb_players);
        }
        g_hash_table_foreach(lobbies, (GHFunc) addline, NULL);
        respond(i, tmp);
       }
       else
       {
        respond(i, "/error This command is only available in a the main chatroom.\n");
       }
      }
      else if(strncmp(buffer, "/nick ", 6)==0)
      {
       buffer[len-1]=0;

       gboolean has_name(int *k, player *p, char *name)
       {
         return (strncmp(p->name, name, strlen(name))==0);
       }

       if(g_hash_table_find(players, (GHRFunc)has_name, buffer+6) != NULL
          || strncmp(buffer+6, "server", 6)*strncmp(buffer+6, "player_", 7)==0 || strlen(buffer+6)<2)
       {
        sprintf(tmp, "/error Username %s is invalid or already taken.\n", buffer+6);
        respond(i, tmp);
       }
       else
       {
        sprintf(tmp, "%s", cur_player->name);
        sscanf(buffer+6, "%63s", cur_player->name);
        sprintf(buffer, "/change %s %s\n", tmp, cur_player->name);
        broadcast(buffer, cur_player, true);
       }
      }
      else if(strncmp(buffer, "/crash", 6)==0)
      {
       broadcast("/msg server Shutting down the server...\n", NULL, true);
       printf("Stopping the server...\n");
       for(j=0; j<fdmax; j++) if(FD_ISSET(j,&master_set)) close(j);
       g_hash_table_destroy(players);
       g_hash_table_destroy(obs);
       g_hash_table_destroy(lobbies);
       exit(0);
      }
      else if(strncmp(buffer, "/", 1)==0)
      {
       buffer[len-1]=0;
       sprintf(tmp, "/error Unrecognized command : %s\n", buffer);
       respond(i, tmp);
      }
      else
      {
       buffer[len]=0;
       sprintf(tmp, "/msg %s %s", cur_player->name, buffer);
       broadcast(tmp, cur_player, true);
      }
     }
    }
   }
 }
}

void usage()
{
  printf("Usage : %s [options]\n\n", pname);
  printf("  A server for a network naval battle game.\n");
  printf("  Possible options for this command are :\n");
  printf("    -h, --help : display this help message\n");
  printf("    -p, --port : set game port number (defaults to 1337)\n");
  printf("    -o, --observer : set observer port number (defaults to 1984)\n");
  printf("    -b, --bind : server bind address\n");
  exit(0);
}

void clerror(char *msg, char *more)
{
 printf("%s %s\n", msg, more);
 usage();
}

int main(int argc, char** argv)
{
 int i;
 char *s;

 pname = argv[0];

 for(i=1; i<argc; i++)
 {
  s=argv[i];
  if(strcmp(s, "-h")*strcmp(s,"--help") == 0)
  {
   usage();
  }
  else if(strcmp(s, "-p")*strcmp(s,"--port") == 0)
  {
    if(++i >= argc) clerror("Missing port number after option", s);
    game_port = (unsigned short) atoi(argv[i]);
    if(game_port <= 1024 || game_port > 65534) clerror("Invalid port number provided for option", s);
  }
  else if(strcmp(s, "-o")*strcmp(s,"--observer") == 0)
  {
    if(++i >= argc) clerror("Missing port number after option", s);
    observer_port = (unsigned short) atoi(argv[i]);
    if(observer_port <= 1024 || observer_port > 65534) clerror("Invalid port number provided for option", s);
  }
  else
  {
   clerror("Unknown command line option:", s);
  }
 }

 server_loop();

 return 0;
}
